/* 
 * pLinguaCore: A JAVA library for Membrane Computing
 *              http://www.p-lingua.org
 *
 * Copyright (C) 2009  Research Group on Natural Computing
 *                     http://www.gcn.us.es
 *                      
 * This file is part of pLinguaCore.
 *
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * pLinguaCore is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pLinguaCore.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.gcn.plinguacore.simulator;

import java.io.PrintStream;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.gcn.plinguacore.util.HashMultiSet;
import org.gcn.plinguacore.util.MultiSet;
import org.gcn.plinguacore.util.Pair;
import org.gcn.plinguacore.util.PlinguaCoreException;
import org.gcn.plinguacore.util.psystem.Configuration;
import org.gcn.plinguacore.util.psystem.Psystem;
import org.gcn.plinguacore.util.psystem.cellLike.CellLikeConfiguration;
import org.gcn.plinguacore.util.psystem.cellLike.membrane.CellLikeMembrane;
import org.gcn.plinguacore.util.psystem.cellLike.membrane.CellLikeSkinMembrane;
import org.gcn.plinguacore.util.psystem.membrane.ChangeableMembrane;
import org.gcn.plinguacore.util.psystem.membrane.Membrane;
import org.gcn.plinguacore.util.psystem.membrane.MembraneStructure;
import org.gcn.plinguacore.util.psystem.rule.IRule;
import org.gcn.plinguacore.util.psystem.rule.checkRule.CheckRule;
import org.gcn.plinguacore.util.psystem.rule.checkRule.specificCheckRule.NoEvolution;



/**
 * An abstract class for simulators
 * 
 *  @author Research Group on Natural Computing (http://www.gcn.us.es)
 * 
 */

public abstract class AbstractSimulator implements ISimulator {

	private transient Thread thread;
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 8223592697775323152L;
	protected Configuration currentConfig = null;
	private Configuration stableCurrentConfig = null;

	private boolean time = false;

	private transient PrintStream infoChannel = System.out;

	private int verbosity = 3;


	private Date date = null;

	
	
	protected transient Psystem psystem;

	
	
	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#setPsystem(org.gcn.plinguacore.util.psystem.Psystem)
	 */
	@Override
	public void setPsystem(Psystem psystem){
		if (psystem == null)
			throw new NullPointerException("Psystem argument shouldn't be null");
		this.psystem = psystem;
		currentConfig = initializePsystemConfiguation(psystem, currentConfig);
		stableCurrentConfig = initializePsystemConfiguation(psystem, stableCurrentConfig);
	}
	private Configuration initializePsystemConfiguation(Psystem psystem,
			Configuration config) {
		if(config==null){
			Configuration configuration =  psystem.getFirstConfiguration();
			return configuration;
		}
		// TODO Auto-generated method stub
		
		return config;
	}
	/**
	 * Creates a simulator which will simulate the P-system passed as argument
	 * 
	 * @param psystem
	 *            The P system to be simulated
	 */
	public AbstractSimulator(Psystem psystem) {
		super();
		this.thread = Thread.currentThread();
		setPsystem(psystem);
		innerReset();
		currentConfig = psystem.getFirstConfiguration();
			
		
	}
	
	
	/**
	 * Resets the simulator to the initial configuration
	 */
	public void reset() {
		currentConfig = psystem.getFirstConfiguration();
		innerReset();
		date = null;
	
	}
	

	
	


	
	
	

	/**
	 * Sets the current configuration of the simulator
	 * @param currentConfig the configuration to be set
	 */
	public void setCurrentConfig(Configuration currentConfig){
		if (currentConfig == null)
			throw new NullPointerException(
					"CurrentConfig argument shouldn't be null");
		this.currentConfig = currentConfig;
		updateStableCurrentConfig();
	}
	
	private void updateStableCurrentConfig(){
		this.stableCurrentConfig =(Configuration)currentConfig.clone(); 
	}
	
	
	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#cleanPreviousConfigurations()
	 */
	public void cleanPreviousConfigurations(){
		
	}
	

	
	
	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#getTime()
	 */
	@Override
	public final double getTime() {

		if (date != null)
			return (((double) ((new Date()).getTime()) - (double) (date
					.getTime())) / 1000);
		return 0;
	}

	/**
	 * Gets the P-system simulated
	 * 
	 * @return The P system simulated
	 */
	public final Psystem getPsystem() {
		return psystem;
	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#supportsStepBack()
	 */
	public boolean supportsStepBack() {
		return false;
	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#supportsAlternateSteps()
	 */
	public boolean supportsAlternateSteps() {
		return false;
	}

	/**
	 * Makes one alternate step of the last one.
	 * 
	 * @return False if there is not alternate configurations. True if a new
	 *         configuration was generated.
	 * @throws UnsupportedOperationException
	 */
	@Override
	public boolean alternateStep() throws UnsupportedOperationException {
		throw new UnsupportedOperationException();

	}

	/**
	 * Counts the number of alternate steps to take
	 * 
	 * @return The number of possible next configurations in the next level of
	 *         the non-deterministic tree
	 * @throws UnsupportedOperationException
	 */
	public int countAlternatives() throws UnsupportedOperationException {
		throw new UnsupportedOperationException();

	}

	/**
	 * Makes one step back.
	 * 
	 * @return False if the current configuration is the initial one. True if a
	 *         new configuration was generated.
	 * @throws UnsupportedOperationException
	 */
	public boolean stepBack() throws UnsupportedOperationException {
		throw new UnsupportedOperationException();

	}

	/*
	 * QUERY: This method has been added to make Simulator instances able to be
	 * stored. For further details, please browse SimulatorFileStream source
	 * code.
	 */

	/**
	 * Gets the info channel
	 * 
	 * @return A PrintStream to write information about the simulation process
	 *         in real-time By defect it is System.out
	 */
	@Override
	public final PrintStream getInfoChannel() {
		return infoChannel;
	}

	/**
	 * Sets the info channel
	 * 
	 * @param infoChannel
	 *            The new PrintStream
	 */
	@Override
	public void setInfoChannel(PrintStream infoChannel) {
		this.infoChannel = infoChannel;
	}

	/**
	 * Makes one step of simulation and generate a new configuration
	 * 
	 * @return false if the current configuration is a halting one. True if a
	 *         new configuration was generated.
	 * @throws PlinguaCoreException
	 *             if a semantic error occurs 
	 */
	public final boolean step() throws PlinguaCoreException{		
		boolean stepTaken =  specificStep();
		/*After taking the step, we need to save the current configuration to avoid not having an stable configuration in case the step couldn't be performed*/
		updateStableCurrentConfig();
		return stepTaken;
	}
	/*
	 * QUERY: This set of methods allows Simulator instances to be stored and
	 * loaded
	 */




	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#runSteps(int)
	 */
	@Override
	public final void runSteps(int steps) throws PlinguaCoreException {
		setThread(Thread.currentThread());
		for (int stepCounter = 0; (!Thread.currentThread().isInterrupted())&&(stepCounter < steps); stepCounter++)
			step();
	}

	private final void runTemporized(ThreadSimulator threadSim, long timeOut) {
		/*
		 * To run a temporized simulator, we delegate on the ThreadSimulator
		 * instance its execution and we monitor the elapsed time
		 */
		synchronized (threadSim) {
			try {
				setThread(threadSim);
				/*We want the thread no to interrupt others when its execution time has elapsed*/
				threadSim.setDaemon(true);
				/*Start the simulator*/
				threadSim.start();
				/*Wait until the thread ends or the time out has elapsed*/

				
				threadSim.wait(timeOut);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				return;
			}
		}
		if (threadSim.isAlive()){
			/*
			 * Once the time has elapsed, the simulation needs to be interrupted
			 * in case it hasn't finished
			 */
			threadSim.tellMonitor();
			threadSim.interrupt();
		}
		/*Stop the thread*/

		/*Wait until the thread dies*/
		//while(!threadSim.isInterrupted());
		/*The current thread is the simulator one again*/
		setThread(Thread.currentThread());

	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#runUntilTimeOut(int)
	 */
	public final void runUntilTimeOut(long timeOut) {
		runTemporized(new ThreadSimulator(this), timeOut);

	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#runUntilTimeOutorSteps(int, int)
	 */
	public final void runUntilTimeOutorSteps(long timeOut, int steps) {
		runTemporized(new StepLimitedThreadSimulator(this, steps), timeOut);
	}

	/**
	 * Takes steps until reaching a halting configuration
	 */

	@Override
	public final void run() throws PlinguaCoreException {
		/*Before running the simulation, we need to set the current thread*/
		setThread(Thread.currentThread());
		while ((!Thread.currentThread().isInterrupted())&&step());	
	}

	private void innerReset(){
		stableCurrentConfig = (Configuration)psystem.getFirstConfiguration().clone();
	}
	
	

	/**
	 * Gets the current configuration
	 * 
	 * @return The current configuration
	 */
	public final Configuration getCurrentConfig() {
		return stableCurrentConfig;
	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#isTimed()
	 */
	@Override
	public boolean isTimed() {
		return time;
	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#setTimed(boolean)
	 */
	@Override
	public void setTimed(boolean time) {
		this.time = time;
	}

	protected final void initDate() {
		if (date == null)
			date = new Date();
	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#getVerbosity()
	 */
	public final int getVerbosity() {
		return verbosity;
	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#setVerbosity(int)
	 */
	public void setVerbosity(int verbosity) {
		this.verbosity = verbosity;
	}

	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#stepsBackAvailable()
	 */
	@Override
	public boolean stepsBackAvailable() {
		// TODO Auto-generated method stub
		return false;
	}
	
	/**
	 * @see org.gcn.plinguacore.simulator.ISimulator#stopThread()
	 */
	@Override
	public void stopThread(){
		if(thread!=null &&thread.isAlive())
			thread.interrupt();
	}
	
	protected void setThread(Thread thread){
		if (thread == null)
			throw new NullPointerException("Thread argument shouldn't be null");
		this.thread = thread;
	}
	
	@Override
	public boolean isFinished() {

		boolean finished = true;
		Iterator<? extends Membrane> it = currentConfig.getMembraneStructure().getAllMembranes()
				.iterator();

		while (it.hasNext() && finished) {
			ChangeableMembrane m =  (ChangeableMembrane)it.next();
			Iterator<IRule> itRules = psystem.getRules().iterator(m.getLabel(),
					m.getCharge());
			while (itRules.hasNext() && finished) {
				if (itRules.next().countExecutions(m,currentConfig.getEnvironment()) > 0)
					finished = false;
			}
		}

		return finished;
	}
	
	
	
	
	
	protected abstract boolean specificStep() throws PlinguaCoreException; 


	

}
